@@ -72,6 +72,9 @@ class Tile:
     # a mapping from dest, source wire to bit coordinates
     pips: Dict[str, Dict[str, Set[Coord]]] = field(default_factory=dict)
     clock_pips: Dict[str, Dict[str, Set[Coord]]] = field(default_factory=dict)
+    # fuses to disable the long wire columns. This is the table 'alonenode[6]' in the vendor file
+    # {dst: ({src}, {bits})}
+    alonenode_6: Dict[str, Tuple[Set[str], Set[Coord]]] = field(default_factory=dict)
     # always-connected dest, src aliases
     aliases: Dict[str, str] = field(default_factory=dict)
     # a mapping from bel type to bel
@@ -160,15 +163,33 @@ def fse_pips(fse, ttyp, table=2, wn=wirenames):
 
     return pips
 
+def fse_alonenode(fse, ttyp, table = 6):
+    pips = {}
+    if 'alonenode' in fse[ttyp].keys():
+        if table in fse[ttyp]['alonenode']:
+            for destid, *tail in fse[ttyp]['alonenode'][table]:
+                fuses = {fuse.fuse_lookup(fse, ttyp, f) for f in unpad(tail[-2:])}
+                srcs = {wirenames.get(srcid, str(srcid)) for srcid in unpad(tail[:-2])}
+                dest = wirenames.get(destid, str(destid))
+                pips[dest] = (srcs, fuses)
+    return pips
+
 # make PLL bels
 def fse_pll(device, fse, ttyp):
     bels = {}
-    # XXX use second grid in order to find PLL ttypes
-    if device in ['GW1N-1',  'GW1NZ-1']:
-        if ttyp == 89:
-            bel = bels.setdefault('RPLLB', Bel())
-        else:
+    if device in {'GW1N-1',  'GW1NZ-1'}:
+        if ttyp == 88:
             bel = bels.setdefault('RPLLA', Bel())
+        elif ttyp == 89:
+            bel = bels.setdefault('RPLLB', Bel())
+    elif device in {'GW1NS-4'}:
+        if ttyp in {88, 89}:
+            bel = bels.setdefault('PLLVR', Bel())
+    elif device in {'GW1N-9C'}:
+        if ttyp in {86, 87}:
+            bel = bels.setdefault('RPLLA', Bel())
+        elif ttyp in {74, 75, 76, 77, 78, 79}:
+            bel = bels.setdefault('RPLLB', Bel())
     return bels
 
 # add the ALU mode
@@ -354,6 +375,7 @@ _known_logic_tables = {
             13: 'BSRAM',
             14: 'DSP',
             15: 'PLL',
+            59: 'CFG',
             62: 'USB',
         }
 
@@ -381,6 +403,7 @@ _known_tables = {
             53: 'DLLDEL0',
             54: 'DLLDEL1',
             56: 'DLL0',
+            60: 'CFG',
             64: 'USB',
             66: 'EFLASH',
             68: 'ADC',
@@ -425,7 +448,18 @@ _pll_loc = {
  'GW1NZ-1':
    {'TRPLL0CLK0': (0, 17, 'F4'), 'TRPLL0CLK1': (0, 17, 'F5'),
     'TRPLL0CLK2': (0, 17, 'F6'), 'TRPLL0CLK3': (0, 17, 'F7'), },
+ 'GW1NS-4':
+   {'TLPLL0CLK0': (0, 27, 'F4'), 'TLPLL0CLK1': (0, 27, 'F7'),
+    'TLPLL0CLK2': (0, 27, 'F6'), 'TLPLL0CLK3': (0, 27, 'F5'),
+    'TRPLL0CLK0': (0, 36, 'F4'), 'TRPLL0CLK1': (0, 36, 'F7'),
+    'TRPLL0CLK2': (0, 36, 'F6'), 'TRPLL0CLK3': (0, 36, 'F5'), },
+ 'GW1N-9C':
+   {'TLPLL0CLK0': (9, 2, 'F4'), 'TLPLL0CLK1': (9, 2, 'F7'),
+    'TLPLL0CLK2': (9, 2, 'F5'), 'TLPLL0CLK3': (9, 2, 'F6'),
+    'TRPLL0CLK0': (9, 44, 'F4'), 'TRPLL0CLK1': (9, 44, 'F7'),
+    'TRPLL0CLK2': (9, 44, 'F5'), 'TRPLL0CLK3': (9, 44, 'F6'), },
 }
+
 def fse_create_pll_clock_aliases(db, device):
     # we know exactly where the PLL is and therefore know which aliases to create
     for row in range(db.rows):
@@ -433,7 +467,7 @@ def fse_create_pll_clock_aliases(db, device):
             for w_dst, w_srcs in db.grid[row][col].clock_pips.items():
                 for w_src in w_srcs.keys():
                     # XXX
-                    if device in {'GW1N-1', 'GW1NZ-1'}:
+                    if device in {'GW1N-1', 'GW1NZ-1', 'GW1NS-4', 'GW1N-9C'}:
                         if w_src in _pll_loc[device].keys():
                             db.aliases[(row, col, w_src)] = _pll_loc[device][w_src]
 
@@ -447,12 +481,13 @@ def from_fse(device, fse):
         tile = Tile(w, h, ttyp)
         tile.pips = fse_pips(fse, ttyp, 2, wirenames)
         tile.clock_pips = fse_pips(fse, ttyp, 38, clknames)
+        tile.alonenode_6 = fse_alonenode(fse, ttyp, 6)
         if 5 in fse[ttyp]['shortval']:
             tile.bels = fse_luts(fse, ttyp)
         if 51 in fse[ttyp]['shortval']:
             tile.bels = fse_osc(device, fse, ttyp)
-        # XXX GW1N(Z)-1 for now
-        if ttyp in [88, 89]:
+        # XXX GW1N(Z)-1 and GW1NS-4 for now
+        if ttyp in [74, 75, 76, 77, 78, 79, 86, 87, 88, 89]:
             tile.bels = fse_pll(device, fse, ttyp)
         tiles[ttyp] = tile
 
@@ -598,13 +633,14 @@ _pll_inputs = [(5, 'CLKFB'), (6, 'FBDSEL0'), (7, 'FBDSEL1'), (8, 'FBDSEL2'), (9,
                (12, 'IDSEL0'), (13, 'IDSEL1'), (14, 'IDSEL2'), (15, 'IDSEL3'), (16, 'IDSEL4'),
                (17, 'IDSEL5'),
                (18, 'ODSEL0'), (19, 'ODSEL1'), (20, 'ODSEL2'), (21, 'ODSEL3'), (22, 'ODSEL4'),
+               (23, 'ODSEL5'), (0, 'RESET'), (1, 'RESET_P'),
                (24, 'PSDA0'), (25, 'PSDA1'), (26, 'PSDA2'), (27, 'PSDA3'),
                (28, 'DUTYDA0'), (29, 'DUTYDA1'), (30, 'DUTYDA2'), (31, 'DUTYDA3'),
                (32, 'FDLY0'), (33, 'FDLY1'), (34, 'FDLY2'), (35, 'FDLY3')]
 _pll_outputs = [(0, 'CLKOUT'), (1, 'LOCK'), (2, 'CLKOUTP'), (3, 'CLKOUTD'), (4, 'CLKOUTD3')]
-def dat_portmap(dat, dev):
-    for row in dev.grid:
-        for tile in row:
+def dat_portmap(dat, dev, device):
+    for row, row_dat in enumerate(dev.grid):
+        for col, tile in enumerate(row_dat):
             for name, bel in tile.bels.items():
                 if bel.portmap: continue
                 if name.startswith("IOB"):
@@ -632,20 +668,71 @@ def dat_portmap(dat, dev):
                         tx = wirenames[dat[f'Iologic{pin}In'][27]]
                         bel.portmap['TX'] = tx
                 elif name == 'RPLLA':
+                    # The PllInDlt table seems to indicate in which cell the
+                    # inputs are actually located.
+                    offx = 1
+                    if device in {'GW1N-9C'}:
+                        # two mirrored PLLs
+                        if col > dat['center'][1] - 1:
+                            offx = -1
                     for idx, nam in _pll_inputs:
                         wire = wirenames[dat['PllIn'][idx]]
-                        bel.portmap[nam] = wire
+                        off = dat['PllInDlt'][idx] * offx
+                        if off == 0:
+                            bel.portmap[nam] = wire
+                        else:
+                            # not our cell, make an alias
+                            bel.portmap[nam] = f'rPLL{nam}{wire}'
+                            dev.aliases[row, col, f'rPLL{nam}{wire}'] = (row, col + off, wire)
                     for idx, nam in _pll_outputs:
                         wire = wirenames[dat['PllOut'][idx]]
+                        off = dat['PllOutDlt'][idx] * offx
+                        if off == 0:
+                            bel.portmap[nam] = wire
+                        else:
+                            # not our cell, make an alias
+                            bel.portmap[nam] = f'rPLL{nam}{wire}'
+                            dev.aliases[row, col, f'rPLL{nam}{wire}'] = (row, col + off, wire)
+                    # clock input
+                    nam = 'CLKIN'
+                    wire = wirenames[dat['PllClkin'][1][0]]
+                    off = dat['PllClkin'][1][1] * offx
+                    if off == 0:
+                        bel.portmap[nam] = wire
+                    else:
+                        # not our cell, make an alias
+                        bel.portmap[nam] = f'rPLL{nam}{wire}'
+                        dev.aliases[row, col, f'rPLL{nam}{wire}'] = (row, col + off, wire)
+                elif name == 'PLLVR':
+                    pll_idx = 0
+                    if col != 27:
+                        pll_idx = 1
+                    for idx, nam in _pll_inputs:
+                        pin_row = dat[f'SpecPll{pll_idx}Ins'][idx * 3 + 0]
+                        wire = wirenames[dat[f'SpecPll{pll_idx}Ins'][idx * 3 + 2]]
+                        if pin_row == 1:
+                            bel.portmap[nam] = wire
+                        else:
+                            # some of the PLLVR inputs are in a special cell
+                            # (9, 37), here we create aliases where the
+                            # destination is the ports of the primitive, but
+                            # you should keep in mind that nextpnr is designed
+                            # so that it will not use such aliases. They have
+                            # to be taken care of separately.
+                            bel.portmap[nam] = f'PLLVR{nam}{wire}'
+                            dev.aliases[row, col, f'PLLVR{nam}{wire}'] = (9, 37, wire)
+                    for idx, nam in _pll_outputs:
+                        wire = wirenames[dat[f'SpecPll{pll_idx}Outs'][idx * 3 + 2]]
                         bel.portmap[nam] = wire
                     bel.portmap['CLKIN'] = wirenames[124];
-                elif name == 'RPLLB':
-                    reset = wirenames[dat['PllIn'][0]]
-                    bel.portmap['RESET'] = reset
-                    reset_p = wirenames[dat['PllIn'][1]]
-                    bel.portmap['RESET_P'] = reset_p
-                    odsel5 = wirenames[dat['PllIn'][23]]
-                    bel.portmap['ODSEL5'] = odsel5
+                    reset = wirenames[dat[f'SpecPll{pll_idx}Ins'][0 + 2]]
+                    # VREN pin is placed in another cell
+                    if pll_idx == 0:
+                        vren = 'D0'
+                    else:
+                        vren = 'B0'
+                    bel.portmap['VREN'] = f'PLLVRV{vren}'
+                    dev.aliases[row, col, f'PLLVRV{vren}'] = (0, 37, vren)
 
 def dat_aliases(dat, dev):
     for row in dev.grid:
