@@ -33,9 +33,23 @@ def sanitize_name(name):
         return retname
     return f"\{retname} "
 
+def extra_pll_bels(cell, row, col, num, cellname):
+    # rPLL can occupy several cells, add them depending on the chip
+    offx = 1;
+    if device == 'GW1N-9C':
+        if int(col) > 28:
+            offx = -1
+        for off in [1, 2, 3]:
+            yield ('RPLLB', int(row), int(col) + offx * off, num,
+                cell['parameters'], cell['attributes'], sanitize_name(cellname) + f'B{off}')
+    elif device in {'GW1N-1', 'GW1NZ-1'}:
+        for off in [1]:
+            yield ('RPLLB', int(row), int(col) + offx * off, num,
+                cell['parameters'], cell['attributes'], sanitize_name(cellname) + f'B{off}')
+
 def get_bels(data):
     later = []
-    belre = re.compile(r"R(\d+)C(\d+)_(?:GSR|SLICE|IOB|MUX2_LUT5|MUX2_LUT6|MUX2_LUT7|MUX2_LUT8|ODDR|OSC[ZFH]?|BUFS|RAMW|RPLL[AB])(\w*)")
+    belre = re.compile(r"R(\d+)C(\d+)_(?:GSR|SLICE|IOB|MUX2_LUT5|MUX2_LUT6|MUX2_LUT7|MUX2_LUT8|ODDR|OSC[ZFH]?|BUFS|RAMW|rPLL|PLLVR)(\w*)")
     for cellname, cell in data['modules']['top']['cells'].items():
         if cell['type'].startswith('DUMMY_') :
             continue
@@ -51,7 +65,11 @@ def get_bels(data):
         if 'DIFF' in cell['attributes'].keys():
             later.append((cellname, cell, row, col, num))
             continue
-        yield (cell['type'], int(row), int(col), num,
+        cell_type = cell['type']
+        if cell_type == 'rPLL':
+            cell_type = 'RPLLA'
+            yield from extra_pll_bels(cell, row, col, num, cellname)
+        yield (cell_type, int(row), int(col), num,
                 cell['parameters'], cell['attributes'], sanitize_name(cellname))
 
     # diff iobs
@@ -180,12 +198,14 @@ _default_pll_internal_attrs = {
 }
 
 # typ - PLL type (RPLL, etc)
-def set_pll_attrs(db, typ, attrs):
+def set_pll_attrs(db, typ, idx, attrs):
     pll_inattrs = add_pll_default_attrs(attrs)
     pll_attrs = _default_pll_internal_attrs.copy()
 
-    if typ not in ['RPLL']:
+    if typ not in {'RPLL', 'PLLVR'}:
         raise Exception(f"PLL type {typ} is not supported for now")
+    if typ == 'PLLVR':
+        pll_attrs[['PLLVCC0', 'PLLVCC1'][idx]] = 'ENABLE'
 
     # parse attrs
     for attr, val in pll_inattrs.items():
@@ -504,11 +524,29 @@ def place(db, tilemap, bels, cst, args):
             for r, c in bits:
                 tile[r][c] = 1
         elif typ.startswith('RPLL'):
-            pll_attrs = set_pll_attrs(db, 'RPLL', parms)
+            pll_attrs = set_pll_attrs(db, 'RPLL', 0,  parms)
+            bits = set()
+            if 'PLL' in db.shortval[tiledata.ttyp].keys():
+                bits = get_shortval_fuses(db, tiledata.ttyp, pll_attrs, 'PLL')
+            #print(typ, bits)
+            for r, c in bits:
+                tile[r][c] = 1
+        elif typ == 'PLLVR':
+            idx = 0
+            if col != 28:
+                idx = 1
+            pll_attrs = set_pll_attrs(db, 'PLLVR', idx, parms)
             bits = get_shortval_fuses(db, tiledata.ttyp, pll_attrs, 'PLL')
             #print(typ, bits)
             for r, c in bits:
                 tile[r][c] = 1
+            # only for 4C, we know exactly where CFG is
+            cfg_type = 51
+            bits = get_shortval_fuses(db, cfg_type, pll_attrs, 'PLL')
+            cfg_tile = tilemap[(0, 37)]
+            for r, c in bits:
+                cfg_tile[r][c] = 1
+
 
         else:
             print("unknown type", typ)
@@ -541,10 +579,31 @@ def place(db, tilemap, bels, cst, args):
         for row, col in bits:
             btile[row][col] = 1
 
+# The vertical columns of long wires can receive a signal from either the upper
+# or the lower end of the column.
+# The default source is the top end of the column, but if optimum routing has
+# resulted in the bottom end of the column being used, the top end must be
+# electrically disconnected by setting special fuses.
+def secure_long_wires(db, tilemap, row, col, src, dest):
+    if device in {"GW1N-1"}:
+        # the column runs across the entire height of the chip from the first to the last row
+        check_row = db.rows
+        fuse_row = 0
+        if row == check_row and dest in {'LT02', 'LT13'}:
+            tiledata = db.grid[fuse_row][col - 1]
+            if dest in tiledata.alonenode_6.keys():
+                tile = tilemap[(fuse_row, col - 1)]
+                _, bits = tiledata.alonenode_6[dest]
+                for row, col in bits:
+                    tile[row][col] = 1
+
+
 def route(db, tilemap, pips):
     for row, col, src, dest in pips:
         tiledata = db.grid[row-1][col-1]
         tile = tilemap[(row-1, col-1)]
+        # short-circuit prevention
+        secure_long_wires(db, tilemap, row, col, src, dest)
 
         try:
             if dest in tiledata.clock_pips:
