@@ -13,6 +13,7 @@ from apycula.attrids import pll_attrids, pll_attrvals
 from apycula.bslib import read_bitstream
 from apycula.wirenames import wirenames
 
+_device = ""
 _pinout = ""
 _packages = {
         'GW1N-1' : 'LQFP144', 'GW1NZ-1' : 'QFN48', 'GW1N-4' : 'PBGA256', 'GW1N-9C' : 'UBGA332',
@@ -89,7 +90,7 @@ def pll_attrs_refine(in_attrs):
         #print(attr, val)
         if attr not in _pll_attrs.keys():
             if attr in ['INSEL', 'FBSEL', 'PWDEN', 'RSTEN', 'CLKOUTDIV3', 'CLKOUTPS']:
-                res.add(f'{attr}={[ name for name, vl in pll_attrvals.items() if vl == val ][0]}')
+                res.add(f'{attr}="{[ name for name, vl in pll_attrvals.items() if vl == val ][0]}"')
             continue
         attr = _pll_attrs[attr]
         if attr in ['CLKOUTP_DLY_STEP', 'CLKOUT_DLY_STEP']:
@@ -161,7 +162,13 @@ _pll_cells = {}
 # GW1N(Z)-1
 def get_pll_A(db, row, col, typ):
     if typ == 'B':
-        col -= 1
+        if _device in {"GW1N-9C"}:
+            if col > 28:
+                col = db.cols - 1
+            else:
+                col = 0
+        else:
+            col -= 1
     return row, col, 'A'
 
 # noiostd --- this is the case when the function is called
@@ -178,8 +185,18 @@ def parse_tile_(db, row, col, tile, default=True, noalias=False, noiostd = True)
     for name, bel in tiledata.bels.items():
         if name.startswith("RPLL"):
             idx = _pll_cells.setdefault(get_pll_A(db, row, col, name[4]), len(_pll_cells))
+            modes = { f'DEVICE="{_device}"' }
+            if 'PLL' in db.shortval[tiledata.ttyp].keys():
+                attrvals = pll_attrs_refine(parse_attrvals(tile, db.logicinfo['PLL'], db.shortval[tiledata.ttyp]['PLL'], pll_attrids))
+                for attrval in attrvals:
+                    modes.add(attrval)
+            if modes:
+                bels[f'{name}{idx}'] = modes
+            continue
+        if name == "PLLVR":
+            idx = _pll_cells.setdefault(get_pll_A(db, row, col, 'A'), len(_pll_cells))
             attrvals = pll_attrs_refine(parse_attrvals(tile, db.logicinfo['PLL'], db.shortval[tiledata.ttyp]['PLL'], pll_attrids))
-            modes = set()
+            modes = { f'DEVICE="{_device}"' }
             for attrval in attrvals:
                 modes.add(attrval)
             if modes:
@@ -500,7 +517,7 @@ def tile2verilog(dbrow, dbcol, bels, pips, clock_pips, mod, cst, db):
         mod.wires.update({srcg, destg})
         mod.assigns.append((destg, srcg))
 
-    belre = re.compile(r"(IOB|LUT|DFF|BANK|CFG|ALU|RAM16|ODDR|OSC[ZFH]?|BUFS|RPLL[AB])(\w*)")
+    belre = re.compile(r"(IOB|LUT|DFF|BANK|CFG|ALU|RAM16|ODDR|OSC[ZFH]?|BUFS|RPLL[AB]|PLLVR)(\w*)")
     if have_iologic(bels):
         bels_items = move_iologic(bels)
     else:
@@ -536,6 +553,15 @@ def tile2verilog(dbrow, dbcol, bels, pips, clock_pips, mod, cst, db):
             portmap = db.grid[dbrow][dbcol].bels[bel[:-1]].portmap
             for port, wname in portmap.items():
                 pll.portmap[port] = f"R{row}C{col}_{wname}"
+        elif typ.startswith("PLLVR"):
+            name = f"PLL_{idx}"
+            pll = mod.primitives.setdefault(name, codegen.Primitive("PLLVR", name))
+            for paramval in flags:
+                param, _, val = paramval.partition('=')
+                pll.params[param] = val
+            portmap = db.grid[dbrow][dbcol].bels[bel[:-1]].portmap
+            for port, wname in portmap.items():
+                pll.portmap[port] = f"R{row}C{col}_{wname}"
         elif typ == "ALU":
             #print(flags)
             kind, = flags # ALU only have one flag
@@ -706,7 +732,7 @@ def fix_pll_ports(pll):
                 pll.portmap.pop(f'{portname}{n}')
 
 def fix_plls(db, mod):
-    for pll_name, pll in [pr for pr in mod.primitives.items() if pr[1].typ == 'rPLL']:
+    for pll_name, pll in [pr for pr in mod.primitives.items() if pr[1].typ in {'rPLL', 'PLLVR'}]:
         if 'INSEL' not in pll.params.keys():
             del mod.primitives[pll_name]
             continue
@@ -725,25 +751,32 @@ def main():
 
     args = parser.parse_args()
 
-    device = args.device
+    global _device
+    _device = args.device
     # For tool integration it is allowed to pass a full part number
-    m = re.match("GW1N(S?)[A-Z]*-(LV|UV|UX)([0-9])C?([A-Z]{2}[0-9]+P?)(C[0-9]/I[0-9])", device)
+    m = re.match("GW1N(S?)[A-Z]*-(LV|UV|UX)([0-9])C?([A-Z]{2}[0-9]+P?)(C[0-9]/I[0-9])", _device)
     if m:
         mods = m.group(1)
         luts = m.group(3)
-        device = f"GW1N{mods}-{luts}"
+        _device = f"GW1N{mods}-{luts}"
 
-    with importlib.resources.open_binary("apycula", f"{device}.pickle") as f:
+    with importlib.resources.open_binary("apycula", f"{_device}.pickle") as f:
         db = pickle.load(f)
 
     global _pinout
-    _pinout = db.pinout[device][_packages[device]]
+    _pinout = db.pinout[_device][_packages[_device]]
 
     bitmap = read_bitstream(args.bitstream)[0]
     bm = chipdb.tile_bitmap(db, bitmap)
     mod = codegen.Module()
     cst = codegen.Constraints()
 
+    # XXX this PLLs have empty main cell
+    if _device in {'GW1N-9C'}:
+        bm_pll = chipdb.tile_bitmap(db, bitmap, empty = True)
+        bm[(9, 0)] = bm_pll[(9, 0)]
+        bm[(9, 46)] = bm_pll[(9, 46)]
+
     for (drow, dcol, dname), (srow, scol, sname) in db.aliases.items():
         src = f"R{srow+1}C{scol+1}_{sname}"
         dest = f"R{drow+1}C{dcol+1}_{dname}"
